# -*- coding: utf-8 -*-


_supported = set()

try:
    import pkg_resources
    # print("loading supported modules...")
    # uncomment to ensure this code indeed only runs once
    _supported = {pkg.key for pkg in pkg_resources.working_set}
except ModuleNotFoundError:
    print("pkg_resources not installed; can't ascertain support of external modules.")



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 10 18:34:24 2020

@author: Dror
"""

import numpy as np

from gensound.utils import num_samples, isnumber

# TODO add capabilities to adapt user-defined functions into curves on the fly
# i.e. x= Curve(lambda k: ...), then passing the lambda into flatten etc.
class Curve():
    def __init__(self, f, duration):
        self.f = f # TODO test with user-supplied f
        self.duration = duration
        
    def multi(self):
        """ converts to 1-d MultiCurve.
        """
        if isinstance(self, MultiCurve):
            return self
        
        return MultiCurve(self)
    
    def flatten(self, sample_rate, inclusive=False):
        """ return a ndarray with the values of self for all sample points
        """
        # implemented for generic f passed into __init__, but in general this is overriden
        return np.asarray(self.f(self.sample_times(sample_rate, inclusive)), dtype=np.float64)
    
    def integral(self, sample_rate):
        """ does the same as flatten, but with the cumulative sum of self
        the implementation here is catch-all, but usually much more efficient
        to override it for specific curves.
        """
        # TODO itertools.accumulate?
        vals = self.flatten(sample_rate, inclusive=True)
        return np.asarray([sum(vals[0:i+1])*i/((i+1)*sample_rate) for i in range(len(vals))], dtype=np.float64)
    
    def endpoint(self):
        """ returns edge value of the curve, which is normally not actually achieved.
        """
        return self.f(self.duration/1000)
    
    def num_samples(self, sample_rate):
        return num_samples(self.duration, sample_rate)
    
    def sample_times(self, sample_rate, inclusive=False):
        if not inclusive:
            return np.linspace(start=0, stop=self.duration/1000, num=self.num_samples(sample_rate), endpoint=False)
        
        return np.linspace(start=0, stop=self.duration/1000, num=self.num_samples(sample_rate)+1, endpoint=True)
    
    # TODO add support for logical or as concat
    # should this be done here or at compoundcurve? imitate signal
    
    def __or__(self, other):
        # concat with number is casted to prolongation of the final value
        # TODO this is one of the few points inconsistent with Signal!
        if isnumber(other):
            other = Constant(value=self.endpoint(), duration=other)
        
        c = CompoundCurve()
        
        if isinstance(self, CompoundCurve):
            c.curves += self.curves
        else:
            c.curves += [self]
        
        if isinstance(other, CompoundCurve):
            c.curves += other.curves
        else:
            c.curves += [other]
        
        return c
    

###### High-level curves #####

class CompoundCurve(Curve):
    """ A concatenation of Curves (similar to Signal/Sequence)
    """
    def __init__(self):
        self.curves = []
    
    def flatten(self, sample_rate):
        return np.concatenate([c.flatten(sample_rate) for c in self.curves])
    
    def integral(self, sample_rate):
        result = self.curves[0].integral(sample_rate)
        # TODO faster implementation?
        for curve in self.curves[1:]:
            # TODO HARD assumption that this function is only useful for frequency curves!
            # therefore the %1, to curb huge numbers
            result = np.concatenate((result[:-1], (result[-1]%1)+curve.integral(sample_rate)))
        
        return result
    
    def endpoint(self):
        return self.curves[-1].endpoint()

    def __getattr__(self, name):
        if name == "duration":
            return sum([c.duration for c in self.curves])
        raise AttributeError

class MultiCurve(Curve):
    """ Multidimensional Curve - one for each dimension (i.e. channel)
    """
    def __init__(self, *curves):
        # TODO implicitly assumes all curves of similar duration
        # will probably crash otherwise
        self.curves = curves
    
    def flatten(self, sample_rate):
        dtype = np.float64 if len(self.curves) == 1 else (np.float64,)*len(self.curves)
        return np.asarray([curve.flatten(sample_rate) for curve in self.curves], dtype=dtype)
    
    def integral(self, sample_rate):
        return np.asarray([curve.integral(sample_rate) for curve in self.curves], dtype=np.float64)
    
    def endpoint(self):
        return np.asarray([curve.endpoint() for curve in self.curves], dtype=np.float64)
    
###### Particular Curves ######

class Constant(Curve):
    def __init__(self, value, duration):
        self.value = value
        self.duration = duration
    
    def flatten(self, sample_rate):
        return np.full(shape=self.num_samples(sample_rate), fill_value=self.value)
    
    def integral(self, sample_rate):
        # TODO maybe slightly different from super.integral, due to start/end conditions
        return np.linspace(start=0, stop=self.value*self.duration/1000,
                           num=self.num_samples(sample_rate)+1, endpoint=True)
    
    def endpoint(self):
        return self.value
    
class Line(Curve):
    def __init__(self, begin, end, duration):
        self.begin = begin
        self.end = end
        self.duration = duration
    
    def flatten(self, sample_rate):
        return np.linspace(start=self.begin, stop=self.end, num=self.num_samples(sample_rate), endpoint=False)
    # TODO class method of computing time ruler, or maybe length
    def integral(self, sample_rate):
        # at^/2+ bt = (at/2+b)*t
        return np.linspace(start=self.begin, stop=(self.end-self.begin)/2+self.begin,
                           num=self.num_samples(sample_rate)+1, endpoint=True) \
            * np.linspace(start=0, stop=self.duration/1000,
                                              num=self.num_samples(sample_rate)+1,
                                              endpoint=True)
    
    def endpoint(self):
        return self.end

class Logistic(Curve): # Sigmoid
    def __init__(self, begin, end, duration):
        self.begin = begin
        self.end = end
        self.duration = duration
        
        self.L = self.end - self.begin # height
        self.T = self.begin # elevation
        self.x0 = self.duration/2000 # midpoint
        self.k = 10*1000/self.duration # steepness
        # TODO parametrize steepness too
        
    
    def flatten(self, sample_rate):
        # 1 / (1 + e^(-k(x-x0)))
        time = self.sample_times(sample_rate)
        return self.L/(1+np.e**(-self.k*(time - self.x0))) + self.T
    
    def integral(self, sample_rate):
        # L/k * ln(1 + e^(k(x-x0))) + Tx
        # TODO faster implementation?
        time = self.sample_times(sample_rate, inclusive=True)
        return (self.L/self.k) * np.log(1 + np.e**(self.k*(time - self.x0))) + self.T * time
    
    def endpoint(self):
        return self.end

class SineCurve(Curve):
    def __init__(self, frequency, depth, baseline, duration):
        self.frequency = frequency
        self.depth = depth
        self.baseline = baseline
        self.duration = duration
    
    def flatten(self, sample_rate):
        return self.depth*np.sin(2*np.pi*self.frequency*self.sample_times(sample_rate)) + self.baseline
    
    def integral(self, sample_rate):
        # sin (x-pi/2) + 1
        return self.depth*np.sin(2*np.pi*self.frequency*self.sample_times(sample_rate, inclusive=True) - np.pi/2) + self.depth + self.baseline*self.sample_times(sample_rate,inclusive=True)
    
    def endpoint(self):
        return 0 # TODO may be troublesome since sample rate is needed
        #raise NotImplementedError
    
class Log(Curve):
    def __init__(self, max, duration, midpoint=-3):
        self.max = max
        self.duration = duration
        self.midpoint = midpoint
        self.curvature = -midpoint/np.log(2)
        assert midpoint < 0
        assert max > 0
    
    def flatten(self, sample_rate):
        return (np.log(self.sample_times(sample_rate)) - np.log(self.duration/1000))*self.curvature + self.max
    
    def integral(self, sample_rate):
        # TODO is this necessary?
        raise TypeError("Log curves should not be used as phases.")
    
    def endpoint(self):
        return self.max




















# -*- coding: utf-8 -*-
"""
Created on Sat Aug 10 09:24:24 2019

@author: Dror
"""

import numpy as np
from gensound.utils import isnumber

midA = 440
octave = 2

semitone = np.power(octave, 1/12)
cent = np.power(semitone, 1/100)

logSemitone = lambda k: np.log(k)/np.log(semitone)

midC = lambda s: midA*semitone**(-9+s)

def freq_to_pitch(freq):
    A0 = 27.5 # lowest on piano?
    if freq < A0:
        return "-"
    semitones_above_A0 = logSemitone(freq/A0)
    closest_pitch = int(round(semitones_above_A0))
    #breakpoint()
    divergence = semitones_above_A0 - closest_pitch
    
    octave = (closest_pitch + 9) // 12
    named_pitch = ["A","A#","B","C","C#","D","D#","E","F","F#","G","G#"][closest_pitch % 12]
    
    return named_pitch + str(octave) + (" " + ("+" if divergence > 0 else "") + str(int(round(divergence*100))) if round(divergence,2) != 0 else "")

def str_to_freq(f): # this is a hack, better use regex or something else
    """ 'C##4+35' middle C## plus 35 cents
    'A' A4 (octave implied)
    """
    cents = 0
    if "+" in f:
        cents = int(f.split("+")[-1])
        f = f.split("+")[0]
    elif "-" in f:
        cents = - int(f.split("-")[-1])
        f = f.split("-")[0]
    
    semi = {"C":0, "D":2, "E":4, "F":5, "G":7, "A":9, "B":11}[f[0]]
    f = f[1:]
    
    while(len(f) > 0 and f[0] in ("#", "b")):
        semi += 1 if f[0] == "#" else -1
        f = f[1:]
    
    if len(f) > 0:
        semi += 12*(int(f)-4) # octave
    
    return midC(semi+cents/100)
        

# TODO make this accessible and modifiable by user
read_freq = lambda f: (f if isnumber(f) else (str_to_freq(f) if isinstance(f, str) else f))
























# -*- coding: utf-8 -*-

from .signals import *
from .transforms import *
from .audio import Audio
from .io import play_Audio, export_WAV
from .musicTheory import midC


mix = Signal.mix
concat = Signal.concat



import os

kushaura = os.path.join(os.path.dirname(__file__), "data/Kushaura_sketch.wav")




# TODO perhaps expose everything inside signals.py and transforms.py,
# leaving amplifiers/filters/effects to be dragged out of their individual files



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  5 19:04:56 2019

@author: Dror
"""

import numpy as np
from gensound.audio import Audio

from gensound.musicTheory import freq_to_pitch

def RMS(audio, start, end):
    return (np.sum(audio.audio[:,start:end]**2) / (audio.num_channels*(end-start)))**0.5


def DFT(audio, N, start=0):
    # TODO only supports mono
    return [(sum([audio.audio[0, start+n]*np.cos(2*np.pi*n*m/N) for n in range(0,N)]),
    -sum([audio.audio[0, start+n]*np.sin(2*np.pi*n*m/N) for n in range(0,N)])) for m in range(0,N)]
    
def DFT_window(audio, N, start=0):
    """ use a window function on the signal to reduce leaking """
    triangle_window = lambda n: 2*n/N if n < (1+N/2) else 2 - 2*n/N
    Hanning_window = lambda n: 0.5 - 0.5*np.cos(2*np.pi*n/N)
    Hamming_window = lambda n: 0.54 - 0.46*np.cos(2*np.pi*n/N)
    
    window = triangle_window
    # TODO test this
    return [(sum([window(n)*audio.audio[0, start+n]*np.cos(2*np.pi*n*m/N) for n in range(0,N)]),
    -sum([window(n)*audio.audio[0, start+n]*np.sin(2*np.pi*n*m/N) for n in range(0,N)])) for m in range(0,N)]

def freq_report(audio, N, sample_rate, start=0):
    FT = DFT(audio, N, start)
    freqs = []
    
    for i, X in enumerate(FT):
        freq = i*sample_rate/N
        pitch = freq_to_pitch(freq)
        
        power = X[0]**2 + X[1]**2
        mag = power**0.5
        
        phase = np.arctan(X[1]/X[0])
        
        freqs.append({
            "frequency" : round(freq, 2),
            "pitch"     : pitch,
            "magnitude" : round(mag, 2),
            "phase"     : round(phase, 2),
            })
    
    return freqs
    

def iDFT(freqs, sample_rate):
    """Freqs is list of [Xreal, Ximag]
    """
    
    N = len(freqs)
    
    xs = []
    
    for n in range(N):
        f = 2*np.pi*n/N
        x = sum([real*np.cos(f*m) + imag*np.sin(f*m) for m, (real, imag) in enumerate(freqs)])
        # TODO make sure that imag*cos and real*sin really do cancel each other
        
        xs.append(x / N)
    
    audio = Audio(sample_rate)
    audio.from_array(xs)
    
    return audio

############
    
def DFT2(samples):
    N = len(samples)
    return [sum([samples[n]*np.e**(-1j*2*np.pi*m*n/N) for n in range(N)]) for m in range(N)]

def iDFT2(freqs):
    N = len(freqs)
    return [sum([freqs[n]*np.e**(1j*2*np.pi*m*n/N) for n in range(N)])/N for m in range(N)]





































# -*- coding: utf-8 -*-
"""
Created on Fri Aug 23 19:03:28 2019

@author: Dror

This file contains the entire interface of the library with the OS.
This includes reading and writing files, and real-time playback.

All of these functionalities should be enclosed in this file,
so as to facilitate cross-platform functionality in the future.

This supports the following basic functionalities:
    * playback from Audio instance
    * Audio instance to file (whatever format possible)
    * file to Audio instance

There is a class for each supported I/O library, which provides these functions
for that library in particular. When this file is loaded, the desired libraries
are determined from settings.py (TODO).

"""

import wave
import inspect
import time

import numpy as np
import simpleaudio as sa
from gensound.settings import _supported

# we start with the support for the various libraries,
# put the exported functionality down below

class _IO_wave:
    is_supported = True # python built-in
    
    @staticmethod
    def export_WAV(filename, audio):
        file = wave.open(filename, "wb")
        file.setnchannels(audio.num_channels)
        file.setsampwidth(audio.byte_width)
        file.setframerate(audio.sample_rate)
        file.setnframes(audio.length)
    
        file.writeframes(audio.buffer)

class _IO_SA:
    is_supported = "simpleaudio" in _supported
    
    @staticmethod
    def playback(audio, is_wait=True): # plays Audio instance
        #audio.buffer = audio.buffer.copy(order='C')
        play_obj = sa.play_buffer(audio.buffer,
                                  num_channels=audio.num_channels,
                                  bytes_per_sample=audio.byte_width, # TODO should Audio store this?
                                  sample_rate=audio.sample_rate)
        
        if not is_wait or audio.audio.shape[1] > 3*10**5:
            input("Type something to quit.")
            play_obj.stop()
        else:
            play_obj.wait_done()


def export_test(audio, func=None, byte_width=2, max_amplitude=1):
    if not hasattr(audio, "buffer"):
        audio._prepare_buffer(byte_width, max_amplitude)
    
    assert inspect.stack()[1].function == func.__name__, "logging names don't match!!!"
    
    timestamp = time.strftime('%Y-%b-%d_%H%M', time.localtime())
    export_WAV("../output/export_{timestamp}_{caller.function}.wav".format( 
                   timestamp = timestamp,
                   caller = inspect.stack()[1]),
                audio)
    
    if func is None:
        return
    
    with open("../output/_export_log.txt", "a") as file:
        # TODO make sure not to overwrite
        file.write("++++++ {} ++++++\n------ {} ------\n\n{}\n\n".format(
                                                               func.__name__,
                                                               timestamp,
                                                               inspect.getsource(func)))

def play_WAV(filename="", is_wait=False):
    wav = sa.WaveObject.from_wave_file(filename)
    play = wav.play()
    
    if not is_wait:
        input("Type something to quit.")
        play.stop()
    else:
        play.wait_done()


def WAV_to_Audio(filename=""):
    wav = sa.WaveObject.from_wave_file(filename)
    # TODO support int24
    buffer = np.frombuffer(wav.audio_data, [np.uint8,np.int16,None,np.int32][wav.bytes_per_sample-1]).astype(np.float64)
    
    buffer = np.reshape(buffer,
                        newshape=(wav.num_channels,
                                  int(len(buffer)/wav.num_channels))[::-1]).T.copy(order='C')
    buffer /= 2**(8*wav.bytes_per_sample-1)
    # TODO the above needs some consideration
    # should we convert to float immediately, or wait till the last minute?
    # is this the right place to convert?
    # and should this be normalized in some way in relation to the synthesized signals?
    from gensound.audio import Audio
    audio = Audio(wav.sample_rate)
    audio.from_array(buffer)
    
    return audio



###################
'''
Here are the exported functionalities.
TODO add logic to figure out supported libraries (also let user customize),
and direct traffic to the correct classes.
also figure out how to navigate different formats
'''

export_WAV = _IO_wave.export_WAV
play_Audio = _IO_SA.playback # should be named 'playback' anywhere but in Audio












# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 18:33:04 2019

@author: Dror

utility functions

"""

from numbers import Number
from collections import Callable
import numpy as np


def isnumber(x): return isinstance(x, Number)
def iscallable(x): return isinstance(x, Callable)
# TODO there is also the controversial built-in callable()

sec = 1000


sample_rates = (8000, 11025, 16000, 22050, 24000, 32000, 44100, 48000, 96000)

def DB_to_Linear(x): return 10**(x/20)
def Linear_to_DB(x): return 20*np.log(x)/np.log(10)




# converts seconds to samples
# later aliased as a Signal class function
# interprets self.duration as either samples or miliseconds
# i.e. using ints will be affected by sample rate
def num_samples(duration, sample_rate): return duration \
                                            if isinstance(duration, int) \
                                            else int(duration*sample_rate/sec)
# this is both for readability as well as for
# bottlenecking non-safe conversions from durations into samples,
# for better control later

def samples_slice(slc, sample_rate): return slice(
                                                None if slc.start is None
                                                     else num_samples(slc.start, sample_rate),
                                                None if slc.stop is None
                                                     else num_samples(slc.stop, sample_rate),
                                                slc.step)


############### Interpolation

# unoptimized code

# These functions receive 2-d ndarray containing audio (audio_array),
# and a list of float indices, and interpolate the values of
# audio_array in these indices.
# higher order are more demanding computationally but yield less artifacts


def interpolate_nearest_neighbor(audio_array, indices):
    if not isinstance(indices, (list, np.ndarray)):
        indices = [indices]
    return audio_array[:,np.rint(indices).astype(int)]

def first_order_interpolation(audio_array, indices):
    if not isinstance(indices, (list, np.ndarray)):
        indices = [indices]
    
    indices = np.array(indices)
    
    flr = np.floor(indices).astype(int)
    ceil = np.ceil(indices).astype(int) # if it happens that flr = ceil, no biggy!
    
    coef = indices - flr # in [0,1)
    
    return (1-coef)*audio_array[:, flr] + coef*audio_array[:, ceil]

# for 2nd/3rd order this may be a drag, becauseof possible out of bounds indices

def second_order_interpolation(audio_array, indices):
    if not isinstance(indices, (list, np.ndarray)):
        indices = [indices]
    
    indices = np.array(indices)
    
    prev_indices = np.floor(indices).astype(int) # sample just before indices (x[n])
    prev_prev_indices = prev_indices - 1 # samples one before that (x[n-1])
    next_indices = np.ceil(indices).astype(int) # sample immediately after (x[n+1])
    tau = indices - prev_indices # tau = sample_index - n
    
    prev_samples = audio_array[:, prev_indices]
    prev_prev_samples = audio_array[:, prev_prev_indices]
    prev_prev_samples[:,prev_prev_indices < 0] = 0 # since indices may be -1, set these samples to 0
    next_samples = audio_array[:, next_indices] # can't overflow
    
    
    coefs = [(tau-1)*(tau)/2, -(tau-1)*(tau+1), tau*(tau+1)/2]
    
    return coefs[0]*prev_prev_samples + coefs[1]*prev_samples + coefs[2]*next_samples
    ...



## choose interpolation by method (used by Audio.resample as well as some Transform)

def get_interpolation(method):
    """
    Methods: TODO names; also use package constant instead?
        'nearest' - zeroth order interpolation (nearest neighbour)
        'linear' - first order interpolation
        'quadratic' - second order interpolation
        'cubic' - 3rd order - not implemented TODO
    """
    if method == 'nearest':
        return interpolate_nearest_neighbor
    elif method == 'linear':
        return first_order_interpolation
    elif method == 'quadratic':
        return second_order_interpolation
    else:
        raise NotImplementedError("Interpolation method '{}' not supported.".format(method))


###########  I/O related

width_by_coding = {"uint8":1,
                   "int16":2,
                   "int24":3,
                   "int32":4,
                   "float32":4}

def audio_to_bytes(audio, coding="int16"):
    # audio is np.ndarray with dtype flot64
    # strategy is to keep dealing with ndarray w/ multiple channels
    # right until the end, where we shave off bytes and interlace etc.
    assert coding in width_by_coding.keys()
    
    num_channels = audio.shape[0]
    length = audio.shape[1]
    byte_width = width_by_coding[coding]
    
    num_samples = num_channels * length
    total_bytes = num_samples * byte_width
    
    audio = audio.copy(order="C") # leave caller Audio object unaffected
    # Better to write everything 10 times than have a proliferation of
    # partially intersecting subcases
    
    if coding == "uint8": # unsigned
        audio = stretch(audio, byte_width)
        audio += 128
        audio = audio.astype(np.uint8)
        
        buffer = np.zeros(num_samples, dtype=np.uint8)
        
        for i in range(num_channels):
            buffer[i::num_channels] = audio[i]
            
    elif coding == "int16":
        audio = stretch(audio, byte_width)
        audio = audio.astype(np.int16)
        
        buffer = np.zeros(num_samples, dtype=np.int16)
        
        for i in range(num_channels):
            buffer[i::num_channels] = audio[i]
            
    elif coding == "int32": # same treatment as int16
        audio = stretch(audio, byte_width)
        audio = audio.astype(np.int32)
        
        buffer = np.zeros(num_samples, dtype=np.int32)
        
        for i in range(num_channels):
            buffer[i::num_channels] = audio[i]
        
    elif coding == "int24": # no np.int24 so needs manual labor
        audio = stretch(audio, 4) # stretch it to fit 4 bytes, later get rid of LSB
        audio = audio.astype(np.int32)
        
        # TODO superfluous step, merge with the next step sometime
        buffer_temp = np.zeros(num_samples, dtype=np.int32)
        for i in range(num_channels):
            buffer_temp[i::num_channels] = audio[i]
        
        buffer_temp = buffer_temp.tobytes()
        
        buffer = bytearray(total_bytes)
        for i in range(num_samples): # take 3 MSBytes, little endian
            buffer[i*3:i*3+3] = buffer_temp[i*4+1:i*4+4] # TODO this truncates, not rounds
        
    elif coding == "float32": # same byte width as int32 but float, different WAV format
        # don't stretch, leave it in [-1,1]
        audio = audio.astype(np.float32)
        
        buffer = np.zeros(num_samples, dtype=np.float32)
        
        for i in range(num_channels):
            buffer[i::num_channels] = audio[i]
    
    return buffer


def stretch(audio, byte_width):
    """ stretches the samples to cover a range of width 2**bits,
    so we can convert to ints later.
    """
    return audio * (2**(8*byte_width-1) - 1)
    

################

def lambda_to_range(f):
    """ transforms function from convenient lambda format to something usable
    for Pan and Amplitude (i.e. shift-sensitive transforms)
    """
    if not iscallable(f):
        f = lambda k: k
    return lambda length, sample_rate: np.asarray([f(x/sample_rate) for x in range(length)], dtype=np.float64)
    # TODO this does not take sample rate into account!


def freq_to_phase(f):
    """ transforms f: time -> frequency to
    g: time -> momentary phase
    """
    # freq being callable
    phase = np.zeros(int(self.sample_rate*self.duration/1000))
    for i in range(1, len(phase)):
        phase[i] = phase[i-1] + 1/self.sample_rate*freq(i/self.sample_rate)
    return 2*np.pi*phase

























# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 13:32:27 2020

@author: Dror
"""

import numpy as np
from gensound.transforms import Transform
from gensound.filters import OnePoleLPF

class GuitarAmp_Test(Transform):
    """ Works by applying a sigmoid to the signal, squashing it somewhat
    violently.
    Sample response is 1/(1+e^(-harshness*x) - 0.5
    
    Limit affects the maximum amplitude achievable (should be lower than maximum
    amplitude in input signal), and harshness the steepness.
    
    Cutoff is for the low-pass applied later to simulate the cabinet.
    """
    def __init__(self, harshness, cutoff=5000, asymmetric=True):
        assert harshness > 0
        self.harshness = harshness
        self.cutoff = cutoff
        self.asymmetric = asymmetric
    
    def realise(self, audio):
        
        # max_amp = self # refactor so that it can reach theo riginal amplitude level
        # now try with asymmetric clipping, hard on the negative and soft on the positives
        test = "soft-soft"
        if test == "soft-hard":
            audio.audio[audio.audio>1] = 1
            negs = audio.audio < -1
            audio.audio[negs] *= (2-np.abs(audio.audio[negs]))
        elif test == "soft-soft":
            audio.audio[audio.audio>1] = 1
            audio.audio[audio.audio<-1] = -1
            mid = (audio.audio > -1) & (audio.audio < 1)
            audio.audio[mid] *= (2-np.abs(audio.audio[mid]))
        elif self.asymmetric:
            #limit = 1/(1+np.e**(-self.harshness)) - 0.5
            limit = 0.5
            positives = audio.audio > 0
            neg_clips = audio.audio < -limit
            audio.audio[positives] = (1/(1+np.e**(-self.harshness*audio.audio[positives])) - 0.5)
            audio.audio[neg_clips] = -limit
        else:
            audio.audio[:,:] = (1/(1+np.e**(-self.harshness*audio.audio)) - 0.5)
        
        lowpass = OnePoleLPF(cutoff=self.cutoff)
        lowpass.realise(audio)
        # TODO add cutoff








 
















# -*- coding: utf-8 -*-
"""
Created on Mon Oct  5 16:46:09 2020

@author: Dror
"""

import numpy as np

from gensound.transforms import Transform, Convolution

class OneImpulseReverb(Convolution):
    def __init__(self, mix=0.5, num=1000, curve="linear"):
        if curve == "linear":
            self.response = np.linspace(mix, 0, num=num)
        elif curve == "steep":
            self.response = 1/np.linspace(1, num, num=num)
        
        self.response.resize((1, self.response.shape[0]))
        # TODO call super().__init__ instead?
        




# -*- coding: utf-8 -*-
"""
Created on Wed Mar 25 15:42:08 2020

@author: Dror
"""

import numpy as np
from gensound.settings import _supported
from gensound.curve import Curve, Line, Logistic, Constant
from gensound.audio import Audio
from gensound.transforms import Transform
from gensound.utils import lambda_to_range, DB_to_Linear, \
                  isnumber, iscallable, \
                  num_samples, samples_slice

# TODO top-class FIR/IIR/Filter?
# that could include a useful function for debugging that generates the impulse response


######## FIRs

class FIR(Transform):
    """ Implements a general-purpose FIR. Subclasses of this can deal solely with
    computing the desired coefficients by overriding FIR.coefficients,
    leaving the actual application to FIR.realise.
    The implementation here may change in the future, and is not guaranteed to be optimal.
    Possibly several alternative implementations will be included, for learning,
    testing and reference purposes. If more competitive implementation is required,
    it is easy enough to extend.
    """
    def __init__(self, *coefficients): # can override this if coefficients are independent of sample rate
        total = sum(coefficients)
        self.h = [c/total for c in coefficients]
        
    def coefficients(self, sample_rate): # override here if sample rate is needed
        # and just ignore the arguments for init
        return self.h
    
    def _parallel_copies(self, audio):
        """ Makes |h| copies of audio, shifting each by the proper amount
        and multiplying by the appropriate coefficient, then summing.
        """
        h = self.coefficients(audio.sample_rate)
        n = audio.length
        parallel = np.zeros((len(h), audio.num_channels, n+len(h)-1), dtype=np.float64)
        
        for i in range(len(h)):
            parallel[i,:,i:n+i] = h[i]*audio.audio
            
        audio.audio[:,:] = np.sum(parallel, axis=0)[:,:n] # TODO trims the end, how to handle this
    
    def _standing_sum(self, audio):
        """ Sums scaled copies of audio into a single ndarray.
        """
        h = self.coefficients(audio.sample_rate)
        new_audio = np.zeros_like((audio.num_channels, audio.length+len(h)-1))
        # could technically skip this first step
        
        for i in range(len(h)):
            new_audio[:,i:audio.length+i] += h[i]*audio.audio
        
        audio.audio[:,:] = new_audio[:,:audio.length] # trims the tail
    
    def realise(self, audio): # override if you have a particular implementation in mind
        self._parallel_copies(audio)
    
    # TODO maybe add class method to facilitate diagnosis of FIR, frequency/phase responses etc.

class MovingAverage(FIR):
    """ Averager Low Pass FIR, oblivious to sample rate.
    """
    def __init__(self, width):
        self.h = [1/width]*width
    

class LowPassBasic(Transform): # TODO I hear a band pass, and a lousy one too
    def __init__(self, cutoff, width):
        self.cutoff = cutoff
        self.width = width # number of samples in FIR
    
    def coefficients(self, sample_rate):
        n = np.linspace(start=-self.width/2, stop=self.width/2, num=self.width+1, endpoint=True)
        omega = 2*np.pi*self.cutoff/sample_rate
        h = np.sin(omega*n)/np.pi/n
        h[int(self.width/2)] = omega/np.pi
        
        
        blackman = [ 0.42 - 0.5*np.cos(2*np.pi*k/(self.width-1)) + 0.08*np.cos(4*np.pi*k/(self.width-1)) for k in range(self.width)]
        
        return h*n
    
    def realise2(self, audio):
        h = self.coefficients(audio.sample_rate)[::-1] # its symmetric tho
        
        res = np.zeros_like(audio.audio, dtype=np.float64)        
        padded = np.pad(audio.audio, ((0,0),(len(h)-1,0)))
        
        for i in range(audio.audio.shape[1]):
            #TODO faster
            res[:,i] = np.sum(h*padded[:,i:i+len(h)], axis=1)
        
        assert res.shape == audio.audio.shape, "FIR distorted audio shape!"
        audio.audio[:,:] = res
    
    def realise(self, audio):
        h = self.coefficients(audio.sample_rate)#[::-1] # its symmetric tho
        n = audio.length
        parallel = np.zeros((len(h), audio.num_channels, audio.length+len(h)-1), dtype=np.float64)
        for i in range(len(h)):
            parallel[i,:,i:n+i] = h[i]*audio.audio
            
        audio.audio[:,:] = np.sum(parallel, axis=0)[:,:n]




############ IIRs

class IIR(Transform):
    """ General-purpose IIR implementation. Subclasses can deal solely with coefficient selection,
    without worrying about the implementation. Override __init__ or coefficients,
    depending on whether or not the sample rate is relevant (typically is).
    """
    def __init__(self, feedforward, feedback): # override this if coefficients are independent of sample rate
        """ Expects two iterables. Feedback[0] is typically 1."""
        self.b = [c/feedback[0] for c in feedforward]
        self.a = [c/feedback[0] for c in feedback]
    
    def coefficients(self, sample_rate): # override this if sample rate is needed
        return (self.b, self.a)
    
    def realise(self, audio): # naive implementation
        # TODO at least the feed-forward coefficients can be computed en masse
        b, a = self.coefficients(audio.sample_rate)
        x = np.pad(audio.audio, ((0,0),(len(a)-1,0))) # max(len(a),len(b))-1
        y = np.zeros_like(x)
        
        for i in range(len(a), x.shape[1]):
            for n in range(len(b)):
                y[:,i] += b[n]*x[:,i-n]
                
            for m in range(1, len(a)):
                y[:,i] -= a[m]*y[:,i-m]
        
        audio.audio[:,:] = y[:,:audio.length]

class OnePoleLPF(IIR):
    """ Designed after Nigel Redmond.
    https://www.earlevel.com/main/2012/12/15/a-one-pole-filter/
    
    For low pass, with Fc being cutoff/sample_rate,
    use b1 = e^{-2 pi Fc}
    and a0 = 1-b1
    
    6dB/octave
    """
    def __init__(self, cutoff):
        self.cutoff = cutoff
    
    def coefficients(self, sample_rate):
        Fc = self.cutoff / sample_rate
        a = (1, -np.e**(-2*np.pi*Fc))
        b = (1 - a[1], )
        return (b, a)


class OnePoleHPF(IIR):
    """ Designed after Nigel Redmond.
    https://www.earlevel.com/main/2012/12/15/a-one-pole-filter/
    
    Not very effective for not very high cutoffs.
    
    b1 = - e^{-2 pi (0.5-Fc)}
    a0 = 1+b1
    
    6dB/octave
    """
    def __init__(self, cutoff):
        self.cutoff = cutoff
    
    def coefficients(self, sample_rate):
        Fc = self.cutoff / sample_rate
        a = (1, np.e**(-2*np.pi*(0.5-Fc)))
        b = (1 + a[1], )
        return (b, a)




class Butterworth(Transform): # LowPass
    def __init__(self, cutoff):
        assert "scipy" in _supported, "SciPy required for Butterworth"
        self.cutoff = cutoff
    
    def realise(self, audio):
        from scipy.signal import butter,filtfilt
        order = 2
        
        normal_cutoff = self.cutoff*2 / audio.sample_rate
        b, a = butter(order, normal_cutoff, btype='low', analog=False)
        audio.audio[:,:] = filtfilt(b, a, audio.audio)











# -*- coding: utf-8 -*-
"""
Created on Sat Aug 17 21:41:28 2019

@author: Dror
"""

import copy

import numpy as np

from gensound.utils import isnumber, num_samples
from gensound.musicTheory import read_freq

from gensound.transforms import Transform, TransformChain, Amplitude, Slice, Combine, BiTransform
from gensound.curve import Curve
from gensound.audio import Audio
from gensound.io import WAV_to_Audio

#CHANNEL_NAMES = {"L":0, "R":1}


__all__ = ["Signal", "Silence", "Step", "WhiteNoise", "Sine", "Triangle",
           "Square", "Sawtooth", "Raw", "WAV"]

class Signal:
    def __init__(self):
        self.transforms = []
    
    #@abstractmethod ???
    def generate(self, sample_rate):
        """
        this is the part the generates the basic signal building block
        ####should return 2d np.ndarray
        should return 1d np.ndarray, the dims are fixed later...
        not sure if good
        TODO
        new: can also return Audio if there is a need;
        but generally should be np.ndarray
        """
        return np.zeros(shape=(1,0))
    
    def realise(self, sample_rate):
        """ returns Audio instance.
        parses the entire signal tree recursively
        """
        audio = self.generate(sample_rate)
        
        if not isinstance(audio, Audio):
            audio = Audio(sample_rate).from_array(audio)
    
        for transform in self.transforms:
            transform.realise(audio=audio)
            
        return audio
    
    mixdown = realise # TODO consider keeping just 1
    
    def play(self, sample_rate=44100, byte_width=2, max_amplitude=1):
        audio = self.realise(sample_rate)
        audio.play(byte_width)
    
    def export(self, filename, sample_rate=44100, byte_width=2, max_amplitude=1):
        """
        0 < max_amplitude <= 1 implies stretching the amplitudes
        so they would hit absolute value of max_amplitude.
        otherwise, max_amplitude = None implies not to touch the amplitudes
        as given, unless they exceed 1 in which case we shrink everything proportionally.
        TODO not sure this behaviour is optimal
        """
        audio = self.realise(sample_rate)
        audio.to_WAV(filename, byte_width)
        
    
    #####################
    def num_samples(self, sample_rate):
        return num_samples(self.duration, sample_rate)
    
    def sample_times(self, sample_rate):
        stop_time = self.duration/1000 if isinstance(self.duration, float) \
            else self.duration/sample_rate
        return np.linspace(start=0, stop=stop_time, num=self.num_samples(sample_rate), endpoint=False)
    
        
    def copy(self):
        """
        creates an identical signal object.
        """
        # TODO perhaps deepcopy should be overridden to be thinner and faster,
        # since it is performed often
        return copy.deepcopy(self)
    
    @staticmethod
    def concat(*args):
        return Sequence(*args)
    
    @staticmethod
    def mix(args):
        # we really want to use #reduce(, args) instead
        # but the question is what should be the 1st argument?
        return sum(args)
    
    ####### sound operations ########
    
    def _amplitude(self, number):
        assert isnumber(number)
        return self*Amplitude(number)
    
    def _apply(self, transform):
        assert not isinstance(transform, BiTransform), "can't apply BiTransform, use concat."
        
        if transform is None:
            return self
        
        if not isinstance(transform, Transform): # TODO change to if isnumber(transform)
            return self._amplitude(transform)
        
        s = self.copy() # TODO is this needed?
        # it is, so we can reuse the same base signal multiple times
        if isinstance(transform, TransformChain):
            s.transforms.extend(transform.transforms)
        else:
            s.transforms.append(transform)
        return s
    
    def _concat(self, other):
        # TODO why is this before mix?
        # TODO consider enabling for negative other,
        # thus shifting a sequence forward(backward? Chinese) in time
        # TODO possibly better way to implement than using Silence
        if isnumber(other):
            other = Silence(duration=other)
            
        s = Sequence()
        # use isinstance(self, Sequence) instead? more semantic
        # TODO write this as overloading of Sequence operators instead?
        if not self.transforms and isinstance(self, Sequence):
            s.sequence += self.sequence
        else:
            s.sequence += [self]
        
        if isinstance(other, BiTransform): # if concatting BiTransform
            s.sequence[-1] = s.sequence[-1]._apply(other.L)
            s.sequence += [other.R]
            return s
        
        if isinstance(s.sequence[-1], Transform): # if coming out of BiTransform
            t = s.sequence.pop()
        else:
            t = None
        
        if not other.transforms and isinstance(other, Sequence):
            other.sequence[0] = other.sequence[0]._apply(t)
            s.sequence += other.sequence
        else:
            s.sequence += [other._apply(t)]
        
        return s
    
    def _mix(self, other):
        # mixing with constant number is interpreted as DC
        # TODO use @singledispatchmethod for this?
        if isnumber(other):
            other = other*DC(duration=self.duration)
        
        s = Mix()
        
        if not self.transforms and isinstance(self, Mix):
            s.signals += self.signals
        else:
            s.signals += [self]
        
        if not other.transforms and isinstance(other, Mix):
            s.signals += other.signals
        else:
            s.signals += [other]
        
        return s
    
    def _repeat(self, number):
        # repeats this signal several times in a row
        assert isinstance(number, int)
        return Signal.concat(*[self]*number)
    
    def _print_nice(self):
        if isinstance(self, Sequence):
            res = "[{}]".format(" + ".join([str(signal) for signal in self.sequence]))
        elif isinstance(self, Mix):
            res = "({})".format(" + ".join([str(signal) for signal in self.signals]))
        else:
            res = str(type(self).__name__)
        
        res += "" if not self.transforms else "*({})".format(",".join([str(transform) for transform in self.transforms]))
        return res
    
    ####### overloading operators #######
    
    def __str__(self):
        return self._print_nice()
    
    def __pow__(self, other):
        return self._repeat(other)
    
    def __rmul__(self, other):
        """ multiplication from the left is only legal for numbers, not for transforms"""
        return self._amplitude(other)
    
    def __mul__(self, other):
        """
        signals can be multiplied on the right by both floats and transforms
        """
        return self._apply(other)
    
    def __radd__(self, other):
        if other == 0:
            return self
        
        raise TypeError("Signal can only be added to other signals, or to 0.")
    
    def __add__(self, other):
        """
        create empty signal list.
        for each operand, if it is transformed or a single signal,
        append it to the list.
        otherwise, it is an untransformed list, simply extend the signal list.
        """
        return self._mix(other)
    
    def __sub__(self, other):
        return self._mix(-other)
    
    def __neg__(self):
        return self._amplitude(-1.0)
    
    def __or__(self, other):
        """Overloading concat."""
        return self._concat(other)
    
    def __ror__(self, other):
        if other == 0:
            return self
        
        if isnumber(other):
            return Silence(duration=other)._concat(self)
        
        raise TypeError("Signal can only be concatted to other signals, or to 0.")
    
    #########
    @staticmethod
    def __subscripts(arg):
        """ Accepts first arg from __get/setitem__ and expands it to include
        both channel as well as time dimension.
        
        args[1] is relevant only for __setitem__.
        
        if args[0] = int,
            then it is taken to refer to a track, e.g.:
                signal[1] = Sine()
                # channel 1 is now a sine wave
        
        if args[0] = slice of ints,
            then it is taken to refer to tracks, e.g.:
                signal = signal[1::-1] # switches L/R
        
        if args[0] = slice of floats,
            then it is taken to refer to time span only:
                signal[3e3:4e3] *= Reverse()
                # reverse the content in the 3rd second
        
        if args[0] = tuple,
            then args[0][0] refers to channel,
            and args[0][1] refers to timespan:
                signal[1:3,10e3:15e3] *= Fade()
                # Applies fade in time 10s-15s only on channels 1, 2
        
        in addition: if the channel is an int (not slice),
            transform it into an equivalent slice
        """
        assert isinstance(arg, (tuple, int, slice))
        
        if isinstance(arg, tuple):
            assert len(arg) == 2, "improper amount of subscripts"
            assert isinstance(arg[1], slice), "time may only be described as ranges"
            if isinstance(arg[0], int):
                arg = (slice(arg[0], arg[0]+1), arg[1])
            assert isinstance(arg[0], slice)
            return arg
        if isinstance(arg, int):
            return (slice(arg, arg+1), slice(None))
        if isinstance(arg, slice):
            # TODO beware of illegal values
            # if both are None, interpret as channels
            if isinstance(arg.start, (int, type(None))) and isinstance(arg.stop, (int, type(None))):
                # slice of channels
                return (arg, slice(None))
            if isinstance(arg.start, (float, type(None))) and isinstance(arg.stop, (float, type(None))):
                # slice of time
                return (slice(None), arg)
            
            raise TypeError("Slice can't be interpreted: inconsistent types")
        
        raise TypeError("Argument not acceptable as subscript")
    
    def __getitem__(self, *args):
        # TODO deal with out-of-bound values for channels and time
        # for now we enable out-of-bound channels when starting at 0 and audio is mono
        return self.copy()*Slice(*Signal.__subscripts(args[0]))
    
    def __setitem__(self, *args):
        assert isinstance(args[1], Signal)
        # TODO deal with out-of-bound values for channels and time
        # TODO should copy self first?
        # TODO use self.apply instead?
        self.transforms.append(Combine(*Signal.__subscripts(args[0]), args[1]))
    
    # def __getattr__(self, name):
    #     # so we can do signal.L to access left channel (0)
    #     if name in CHANNEL_NAMES:
    #         return self[CHANNEL_NAMES[name]]
        # problems: we need to define self.L differently when its setattr vs. getagttr
        # and also avoid recursion when using setattr
    #     raise AttributeError
    
    def __iter__(self):
        # can't iterate over channels since they're not known until mixdown
        # (transforms affect them during realise)
        # this is to prevent sum(signal) as a way to mix into mono
        # instead of getting into a loop (didn't go deep into that one)
        # 
        raise TypeError("May not iterate over Signal objects, as channel number is unknown prior to mixdown.")

#### other "high-level" signals ##############################3

class Mix(Signal):
    """
    a list of signals to be mixed together
    (AKA "internal node" of the mix tree)
    """
    def __init__(self, *signals):
        super().__init__()
        self.signals = list(signals)
    
    def generate(self, sample_rate):
        audio = Audio(sample_rate)
        
        for signal in self.signals:
            audio += signal.realise(sample_rate)
        
        return audio

class Sequence(Signal):
    """
    a list of signals to be placed one after the other
    """
    def __init__(self, *sequence):
        super().__init__()
        self.sequence = list(sequence)
    
    def generate(self, sample_rate):
        audio = Audio(sample_rate)
        
        for signal in self.sequence:
            audio.concat(signal.realise(sample_rate))
            # TODO assymetric with Mix since we don't overload audio.concat
        
        return audio

#### particular signals #########

class Silence(Signal):
    def __init__(self, duration=5e3):
        super().__init__()
        self.duration = duration
    
    def generate(self, sample_rate):
        return np.zeros(self.num_samples(sample_rate), dtype=np.float64)

class Step(Signal): # Impulse? DC?
    def __init__(self, duration=1):
        super().__init__()
        self.duration = duration
    
    def generate(self, sample_rate):
        return np.ones((self.num_samples(sample_rate),), dtype=np.float64)

DC = Step

## TODO step with frequency

class WhiteNoise(Signal):
    def __init__(self, duration=5e3):
        # TODO add seed argument?
        super().__init__()
        self.duration = duration
    
    def generate(self, sample_rate):
        # TODO this may have non-zero DC!
        return 2*np.random.rand(self.num_samples(sample_rate)) - 1

# TODO add grey, pink noise.


#### simple oscillators


class Sine(Signal): # oscillator? pitch? phaser?
    wave = np.sin
    
    def __init__(self, frequency=220, duration=5e3):
        # TODO add initial phase argument (for completeness; also for potential phase inference)
        super().__init__()
        self.frequency = read_freq(frequency)
        self.duration = duration
        # TODO if frequency is a curve then duration should be derived from it?
        # or ignored?
        
    def generate(self, sample_rate):
        # TODO currently the [:-1] after the integral is needed,
        # otherwise it would be one sample too long. perhaps there is more elegant solution,
        # maybe passing an argument telling it to lose the last sample,
        # or better, having CompoundCurve give the extra argument telling its
        # children NOT to lose the last sample
        if isinstance(self.frequency, Curve):
            return type(self).wave(2*np.pi * self.frequency.integral(sample_rate)[:-1])
        return type(self).wave(2*np.pi * self.frequency * self.sample_times(sample_rate))

class Triangle(Sine): # TODO start at 0, not 1
    wave = lambda phase: 2*np.abs((phase % (2*np.pi) - np.pi))/np.pi - 1
    
class Square(Sine):
    wave = lambda phase: ((phase % (2*np.pi) < np.pi)*2 - 1).astype(np.float64)

class Sawtooth(Sine):
    wave = lambda phase: (phase % (2*np.pi))/np.pi-1

# TODO sweepsine

### raw audio signals

class Raw(Signal):
    cache = {}
    """
    Keep track of when the audio is copied;
    we should probably use a view until we start applying transforms.
    I.e. this object should only keep a view, and on generate it should copy.
    """
    def __init__(self, audio=None):
        super().__init__()
        
        if audio != None:
            if not hasattr(self, "key"):
                self.key = len(Raw.cache)
            
            if self._key() not in Raw.cache:
                Raw.cache[self._key()] = audio
        
    def _key(self): # TODO __key__ ?
        return type(self).__name__ + ":" + str(self.key)
    
    def resample(self, sample_rate=44100, method="quadratic"):
        self.audio._resample(sample_rate, method)
        return self # so the user can type WAV("a.wav").resample(44100)*Transform() etc.
        ...
    
    @property
    def audio(self):
        # return Audio object, not ndarray, to override sample raes TODO see if this makes sense overall
        return Raw.cache[self._key()]#.audio
    
    def generate(self, sample_rate):
        #return np.copy(self.audio.audio)
        return self.audio.audio
    """
    TODO
    ####think about this more. here we're copying the audio data,
    #### but not the audio object. should we copy the audio object instead maybe?
    we're passing the direct audio buffer.
    since this eventually goes to audio.from_Array, in which np.copy is called,
    this SHOULD not cause problems when using the same shell for different copies
    of the same original signal.
    also what happens if self.audio already has a sample rate, and the two don't agree?
    """


class WAV(Raw):
    cache = {}
    """
    TODO perhaps make Raw.cache instead
    (but make it easily extensible for new subclasses of Raw)
    either way, WAV/Raw objects should not contain
    the actual audio!
    just a key that will only be used in generate().
    this way the object is an empty skeleton.
    
    Raw.cache can be a dictionary with types as keys,
    values are actual caches with keys defined by each subclass individually
    """
    
    def __init__(self, filename):
        self.key = filename
        
        audio = None
        
        if self._key() not in Raw.cache:
            audio = WAV_to_Audio(filename)
        
        # TODO copy again? so the cache will be eternally independent?
        super().__init__(audio)
    

























# -*- coding: utf-8 -*-

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='gensound',
      version='0.2',
      author='Dror Chawin',
      author_email='quefumas@gmail.com',
      description='Pythonic audio processing and synthesis library',
      packages=['gensound'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/Quefumas/gensound',
      install_requires=['numpy','simpleaudio'],
      package_data={'gensound': ['data/Kushaura_sketch.wav']},
      )


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 22 20:53:05 2019

@author: Dror

TODO
do we need all these functions as static?
some may be instance functions, some may belong in utils.
"""

import numpy as np
import copy

from gensound.utils import sec

class Audio:
    """Basically a wrapper for a numpy array, representing the signal.
    Shape is (tracks, samples), tracks being >= 1.
    This object is mostly dealt with inside Transform and Signal functions,
    and typically should not be used directly unless you're creating new
    Signals or Transforms.
    Don't override too many operators - this is the nuts and bolts,
    not the pretty facade.
    """
    def __init__(self, sample_rate):
        self.shift = 0
        self.sample_rate = sample_rate
        # TODO get rid of sample_rate argument and have optional audio argument
        # that can be np.ndarray or other array or something
        self.audio = np.zeros((1, 0), dtype=np.float64)
    
    def ensure_2d(self): # TODO should move to utils
        """ makes sure self.audio is 2-dimensional,
        since this is not obvious for mono signals.
        """
        # TODO make sure this is called in a hermetic fashion, everywhere it is needed
        if len(self.audio.shape) == 1:
            self.audio.resize((1, self.audio.shape[0]))
            return False # indicating it wasn't 2d
        return True # the shape was fine all along
        
    
    def from_array(self, array):
        """
        converts np.ndarray to Audio.
        if array is not of type np.float64, converts it implicitly!
        note that this normalizes the values to be within [-1,1]
        """
        if not isinstance(array, np.ndarray):
            array = np.asarray(array, dtype=np.float64)
        
        self.audio = array
        self.ensure_2d()
        
        # TODO should we copy?
        # note that this is called practically everytime we generate() a signal!!!
        self.audio = self.audio.copy(order="C")
        
        return self
    
    def copy(self):
        """
        creates an identical Audio object.
        """
        return copy.deepcopy(self)
    
    ####### getters #######
    
    # always get the info from self.audio itself;
    # we manipulate it directly all the time
    @property
    def num_channels(self):
        assert len(self.audio.shape) == 2
        return self.audio.shape[0]
    
    @property
    def is_mono(self):
        return self.num_channels == 1
    
    @property
    def length(self):
        """ Returns length in samples.
        """
        assert len(self.audio.shape) == 2
        return self.audio.shape[1]
    
    @property
    def duration(self):
        """ Returns duration in ms.
        """
        # TODO this appears to be the only method the uses self.sample_rate
        # consider having it as an argument instead of instance variable
        # reply: Audio always has sample rate, so this should always be available.
        return self.length/self.sample_rate*sec
    
    @property
    def shape(self): # TODO consider doing others like this
        return self.audio.shape
    
    def abs_start(self): return self.shift # in samples only
    def abs_end(self): return self.shift+self.length
    
    ######### Unary manipulations #########
    
    """
    
    first make sure other is appropriate type
    can be scalar (if multiply)
    or np.list, in which case we may need to add channels first
    
    length not restrictive
    
    """
    
    def conform(self, other):
        """
        reshapes self.audio so other.audio may be mixed into it.
        
        ensures other is a 2-d ndarray of similar 1st shape,
        and that self.length >= other.length
        and that self has enough channels
        note that this function has side effects for self.audio!
        
        and unfortunately also for other.
        """
        
        assert isinstance(other, Audio), "Audio.conform can only be used between Audios"
        assert other.is_mono or self.is_mono or other.num_channels == self.num_channels
        
        # conforming channels
        if other.is_mono:
            other.from_mono(self.num_channels)
            # TODO warning: this affects other.
        
        if self.is_mono:
            self.from_mono(other.num_channels)
        
        # conforming lengths TODO better way?
        start = min(self.abs_start(), other.abs_start())
        end = max(self.abs_end(), other.abs_end())
        self.push_forward(self.abs_start() - start)
        self.to_length(end - start)
        
    
    
    ### time manipulations ###
    
    def to_length(self, length):
        """ Ensures self.length is at least length, padding with zeros if needed.
        """
        # TODO this and extend, do we need both?
        if self.length >= length:
            return
        
        self.extend(length - self.length)
    
    def extend(self, how_much):
        """ extends all available channels with zeros """
        self.audio = np.pad(self.audio, ((0,0),(0,how_much)), mode="constant", constant_values=0.0)
    
    def push_forward(self, how_much):
        """ pads the beginning with zeros """
         # TODO sister function which truncates beginning when shift < 0? for use in Combine
        if how_much < 0:
            return
        # pad with nothing before and after the channel dimension
        # pad with how_much before the time dimension, 0 after
        self.audio = np.pad(self.audio, ((0,0),(how_much,0)), mode="constant", constant_values=0.0)
        self.shift -= how_much
    
    def _resample(self, sample_rate, method):
        """ Uses interpolation to change the sample rate of the audio
        while retaining spectral content.
        Don't use this directly; better use Raw().resample instead.
        [TODO does this have to be a supported sample rate?]
        """
        if self.sample_rate == sample_rate:
            return
        from gensound.utils import get_interpolation
        interpolate = get_interpolation(method)
        factor = self.sample_rate / sample_rate
        self.audio = interpolate(self.audio, np.arange(0, self.length-1, factor))
        self.sample_rate = sample_rate
    
    ### channel manipulations ###
    
    def to_mono(self):
        """ mixes all channels down to one. does not scale!
        """
        self.audio = np.sum(self.audio, 0)
    
    def from_mono(self, num_channels):
        """ duplicates a mono channel into various channels.
        does not scale! """
        assert self.is_mono, "Can't call Audio.from_mono() for non-mono Audio."
        if num_channels == 1:
            return
        self.audio = self.audio.copy() # TODO is this wasteful?
        self.audio.resize((num_channels, self.length), refcheck=False)
        self.audio[:,:] = self.audio[0,:]
        assert self.ensure_2d()
    
    def to_channels(self, num_channels):
        """ ensures there are at least num_channels
        """
        shape = (num_channels-self.num_channels, self.length)
        self.audio = np.vstack((self.audio, np.zeros(shape, dtype=np.float64)))
    
    ######## binary operations ###########
    
    def mix(self, other):
        assert isinstance(other, Audio)
        
        self.conform(other)
        self.audio[:,other.abs_start()-self.abs_start():other.abs_start()-self.abs_start()+other.length] += other.audio
        return self
    
    def concat(self, other):
        assert isinstance(other, Audio)
        
        other.shift += self.length
        self.mix(other)
    
    
    
    ######## Overloading Operators ###########
    ## TODO consider overloading __setitem__, __getitem__,
    # so that we can do operations on audio instead of audio.audio
    # (or is it being too wordy here?)
    # also, choosing a single channel (not a slice) can be dealt with in the process,
    # delegating this responsibility away from Signal.__subscripts
    
    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return other.__add__(self)
    
    def __add__(self, other): # do we want to overload this?
        return self.mix(other)
    
    def __mul__(self, other):
        print("convolution, should not happen")
        # TODO this should be reexamined and perhaps moved into Convolution or sth
        # also delegate to a non-overloaded function first
        if isinstance(other, np.ndarray):
            assert len(other.shape) == 1, "can multiply Audio by np.ndarray only for one-dimensional arrays"
            if other.shape[0] > self.length:
                other = other[0:self.length]
            self.audio[:,0:other.shape[0]] *= other
            return
            
        assert isinstance(other, Audio)
        # for multiplying by a float, we multiply the signal instead
        # TODO also does not support with Audios with differing params
        self.conform(other)
        self.audio[:,other.abs_start()-self.abs_start():other.abs_start()-self.abs_start()+other.length] *= other[:,:]
        return self
    
    
    
    
    
    
    
    ######## prepare for mixdown ########
    """
    we have these as static since in Audio.mixdown(), we do not wish
    to affect self.audio.
    
    
    TODO think more about this. fit() is probably desirable for playback,
    but Audio.buffer property and everything related to conversion to bytes
    should probably belong strictly in io.py.
    
    mixdown() shouldn't even exist.
    we can keep fit() and call it when doing playback or export, since it makes
    sense to have it around, like extend, resample etc.
    But export should actually be called from Signal, and do the rest on its own.
    """
    
    def fit(self, max_amplitude=1):
        """
        stretches/squashes the amplitude of the samples to be [-max_amplitude, +max_amplitude]
        for max_amplitude <= 1.
        
        for max_amplitude = None, leave things as they are.
        
        'destructive' function, called only when exporting or playing.s
        """
        assert max_amplitude is None or 0 < max_amplitude
        
        if max_amplitude == None:
            return
        
        max_amp = np.max(np.abs(self.audio))
        
        if max_amp == 0:
            return
        
        self.audio = self.audio * (max_amplitude / max_amp)
        
    
    ####### post-mixdown ########
    
    def _prepare_buffer(self, byte_width=2, max_amplitude=1):
        """ Attaches byte width to Audio object, and converts audio information
        to bytes object. This is only done in preparation for output (file or playback).
        """
        from gensound.utils import audio_to_bytes
        
        self.push_forward(self.shift)
        self.fit(max_amplitude)
        
        self.byte_width = byte_width
        self.buffer = audio_to_bytes(self.audio, ["uint8","int16","int24","int32"][byte_width-1])
        
        
    def play(self, byte_width=2, is_wait=True):
        from gensound.io import play_Audio
        
        self._prepare_buffer(byte_width)
        play_Audio(self, is_wait)
    
    def to_WAV(self, filename, byte_width=2):
        from gensound.io import export_WAV
        
        self._prepare_buffer(byte_width)
        
        # TODO disentangle export_WAV arguments so that they will be given
        # seperately (or in a config dict) rather than in an Audio object,
        # since the latter is not supposed to hold a buffer nor byte width
        export_WAV(filename, self)
    












    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Aug 18 21:01:16 2019

@author: Dror
"""

import numpy as np
from gensound.settings import _supported
from gensound.curve import Curve, Line, Logistic, Constant
from gensound.audio import Audio
from gensound.utils import lambda_to_range, DB_to_Linear, \
                  isnumber, iscallable, \
                  num_samples, samples_slice, sec

__all__ = ["Transform", "Shift", "Extend", "Reverse", "Fade", "CrossFade",
           "Gain", "SineAM", "Limiter", "Mono", "Pan", "Repan", "Downsample",
           "Convolution", "ADSR"]

class Transform:
    """ represents post-processing on some given signal.
    use __init__ to set the transform params,
    and realise for the implementation on the WAV.
    
    realise should directly change the audio of the signal element,
    and thus each transform can be bound to several signals without intereference.
    
    """
    
    def __init__(self):
        # TODO consider using *kwargs and making it copy all attributes to self
        # this would save us many inherited inits simply doing self.duration = duration
        # OTOH could this break something by name collision?
        # also, wouldn't this make the code less clear (note that num_samples() also
        # uses hidden self.duration)
        pass
    
    def __str__(self):
        return str(type(self).__name__)
    
    def __mul__(self, other):
        if isinstance(other, Transform):
            t = TransformChain()
            
            if isinstance(self, TransformChain):
                t.transforms.extend(self.transforms)
            else:
                t.transforms.append(self)
            
            if isinstance(other, TransformChain):
                t.transforms.extend(other.transforms)
            else:
                t.transforms.append(other)
            
            return t
        
        return NotImplemented
    
    def num_samples(self, sample_rate):
        if not hasattr(self, "duration"):
            raise TypeError("transform.duration must be defined to support conversion to samples")
        return num_samples(self.duration, sample_rate)
    
    def realise(self, audio):
        """ here we apply the transformation on the Audio object.
        this should change the object directly, don't return anything."""
        pass

####### High-level transforms ################

class TransformChain(Transform):
    """ This class lets you combine a set of transforms which will then be
    applied in one go to a signal.
    
    realise() is not implemented, since when this is applied to a signal,
    all transforms are applied individually and there is no need for this anymore.
    
    TODO consider realising them together as well; possibly this will look nicer
    when printing the signal tree.
    """
    def __init__(self, *transforms):
        self.transforms = list(transforms)


class BiTransform(Transform):
    """ Bi-directional transform, which may only (?) be used in concatenation
    in between two signals. This is basically a container of two TransformChains,
    It applies the first to the left signal and the second to the right.
    
    Again, no overriding of realise() since this is to be broken up when 
    undergoing the first concatenation (from the left).
    """
    def __init__(self, L, R):
        assert isinstance(L, Transform) and isinstance(R, Transform)
        self.L = L
        self.R = R

####### Slicing Stuff ################

class Slice(Transform):
    """ returns only a specified part of the signal.
    Should not be invoked by the user, rather it is used by
    Signal.__getitem.
    
    # TODO if hell froze over maybe its possible to merge this with Combine
    """
    def __init__(self, channel_slice, time_slice):
        # TODO filter slices (if relevant?)
        self.channel_slice = channel_slice
        self.time_slice = time_slice
    
    def realise(self, audio):
        if audio.is_mono and self.channel_slice.start == 0 and self.channel_slice.stop > 0:
            audio.from_mono(self.channel_slice.stop)
        audio.audio = audio.audio[self.channel_slice,
                                  samples_slice(self.time_slice, audio.sample_rate)]
        assert audio.ensure_2d(), "pbbly channel_slice wasn't a slice, so not ensure_2d()"

class Combine(Transform):
    """ given another Signal as input, realises it and pushes it back
    into the affected signal in the relevant place.
    if the signal to push inside is too small, it will mean a silence gap,
    if too long, its remainder will be mixed into the continuation of the parent signal.
    TODO reevaluate this behaviour. perhaps when there is no specified end to the
    time slice (for example signal[:,5e3:]), then we can extend. otherwise, maybe not.
    also consider that the overflow may happen in the opposite direction (negative shift)
    Should not be invoked by the user, rather it is used by
    Signal.__setitem.
    """
    def __init__(self, channel_slice, time_slice, signal):
        self.channel_slice = channel_slice # slices of container Signal
        self.time_slice = time_slice
        self.signal = signal # inserted Signal
    
    def realise(self, audio):
        # prepare new audio and ensure shift can only be negative
        new_audio = self.signal.realise(audio.sample_rate)
        new_audio.push_forward(new_audio.shift)
        
        # locating correct samples
        sample_slice = samples_slice(self.time_slice, audio.sample_rate)
        start_sample = sample_slice.start if sample_slice.start is not None else 0
        
        # add channels to container in case of out-of-bounds Container channel subscript
        max_channel = max(0, self.channel_slice.start or 0, (self.channel_slice.stop or 0)-1)
        if max_channel >= audio.num_channels:
            audio.to_channels(max_channel+1)
        
        # ensure container long enough in case of overflow
        audio.to_length(start_sample + new_audio.length)
        
        # emptying the sliced area
        audio.audio[self.channel_slice, sample_slice] = 0 # TODO is this needed?
        
        # put inside; recall that new_audio.shift <= 0
        audio.audio[self.channel_slice, start_sample:start_sample+new_audio.length+new_audio.shift] += new_audio.audio[:,-new_audio.shift:]


####### Basic shape stuff ##############

class Shift(Transform):
    """ shifts the signal forward in time."""
    # TODO enable backward
    # TODO doesn't seem to work when its a forward shift for the first signal in the mix
    def __init__(self, duration):
        self.duration = duration
    
    def realise(self, audio):
        audio.shift += self.num_samples(audio.sample_rate)

class Extend(Transform):
    """ adds silence after the signal. needed?
    """
    def __init__(self, duration):
        self.duration = duration
    
    def realise(self, audio):
        audio.extend(self.num_samples(audio.sample_rate))

class Reverse(Transform):
    """
    reverses the signal
    """
    def realise(self, audio):
        audio.audio[:,:] = audio.audio[:,::-1]


######### Level/ampltidue Stuff ###################

class Fade(Transform):
    min_fade = -50
    
    def __init__(self, is_in=True, duration=3e3):
        self.is_in = is_in
        self.duration = duration
    
    def realise(self, audio):
        amp = DB_to_Linear(np.linspace(Fade.min_fade, 0, self.num_samples(audio.sample_rate)))
        #amp = np.linspace(0, 1, self.num_samples(audio.sample_rate))
        # perhaps the fade in should be nonlinear
        # TODO subsciprability problem
        
        if self.is_in:
            audio.audio[:,:len(amp)] *= amp
        else:
            audio.audio[:,-len(amp):] *= amp[::-1]
        
        # TODO in case of fade out, if amp is shorter or longer than audio,
        # care must be taken when multiplying!
        # TODO TEST!!!!!!!!!!!!!
        # TODO I still hear a bump when the playback starts


class CrossFade(BiTransform): # TODO rename to XFade?
    def __init__(self, duration):
        L = Fade(is_in=False, duration=duration)
        R = Fade(duration=duration)*Shift(-duration)
        super().__init__(L, R)
        # needs a lot of tweaking of fade curves and db vs. linear stuff


class Gain(Transform):
    """
    Adds positive/negative gain in dBs to the signal.
    """
    def __init__(self, *dBs):
        self.dBs = dBs
    
    def realise(self, audio):
        # should we make this inherit from amplitude? like Triangle/Sine relation?
        
        # when given a single gain for multiple channels, apply it to all of them
        if len(self.dBs) == 1 and audio.num_channels > 1:
            dBs = self.dBs * audio.num_channels
        else:
            dBs = self.dBs
        
        for (i, dB) in enumerate(dBs):
            if isnumber(dB):
                audio.audio[i,:] *= DB_to_Linear(dB)
            elif isinstance(dB, Curve):
                vals = DB_to_Linear(dB.flatten(audio.sample_rate))
                audio.audio[i,0:dB.num_samples(audio.sample_rate)] *= vals
                audio.audio[i,dB.num_samples(audio.sample_rate):] *= DB_to_Linear(dB.endpoint())
            else:
                raise TypeError("Unsupported amplitude type")

class Amplitude(Transform):
    """ simple increase/decrease of amplitude.
    for constant amplitude, don't use this directly;
    best to just use 0.34 * Signal syntax for example, which reverts to this class.
    use this for more complex amplitude functions
    
    use Gain() to change in dB
    """
    def __init__(self, *amps):
        self.amps = amps
    
    def realise(self, audio):
        # TODO do multiple channels at the same time?
        
        # when given a single gain for multiple channels, apply it to all of them
        if len(self.amps) == 1 and audio.num_channels > 1:
            amps = self.amps * audio.num_channels
        else:
            amps = self.amps
            
        for (i,amp) in enumerate(amps):
            # TODO shouldn't this just affect a copy of audio????
            # above comment was previously for the line:
            # audio.audio = self.size*audio.audio
            if isnumber(amp):
                audio.audio[i,:] *= amp
            elif isinstance(amp, Curve):
                vals = amp.flatten(audio.sample_rate)
                # TODO view or copy?
                # TODO what if curve duration doesn't match signal?
                # or can we have curve duration extracted from signal? automatically matching it
                # should we just stretch the last value of curve?
                # should we define curve.conform?
                # also what if curve is too long?
                audio.audio[i,0:amp.num_samples(audio.sample_rate)] *= vals
                audio.audio[i,amp.num_samples(audio.sample_rate):] *= amp.endpoint()
            else:
                raise TypeError("Unsupported amplitude type")

class SineAM(Transform):
    """
    have the amplitude change according to a sine function over time continuously,
    with given width (size) and frequency
    TODO again the factors should be perhaps logarithimic
    """
    def __init__(self, frequency, size, phase=0):
        self.frequency = frequency
        self.size = size
        self.phase = phase
    
    def realise(self, audio):
        """ audio is Audio instance"""
        assert isinstance(audio, Audio) # TODO remove this after debug hopefully
        
        sin = np.sin(self.phase + self.frequency * \
                     np.linspace(0, audio.duration/sec, audio.length, False) * 2 * np.pi)
        audio *= (sin * self.size + (1-self.size))
        # remember [:] is necessary to retain changes
        


class Limiter(Transform):
    """ Cuts any sample that exceed some amount.
    Amount can be an absolute value amplitude,
    or a ratio/percentage of the current maximum.
    
    minimum params are to limit from below in the range [-min,+min]
    which is just experimental
    """
    def __init__(self, max_amplitude=None, max_ratio=None, max_dB=None,
                       min_amplitude=None, min_ratio=None, min_dB=None):
        """ at most one max can to be non-None, at most one min as well"""
        assert (max_amplitude is not None) + (max_ratio is not None) + (max_dB is not None) <= 1
        assert (min_amplitude is not None) + (min_ratio is not None) + (min_dB is not None) <= 1
        
        #TODO more assertions
        
        if max_dB is not None or min_dB is not None:
            raise NotImplementedError
            # TODO
        
        self.is_min = not (min_amplitude is None and min_ratio is None and min_dB is None)        
        self.is_max = not (max_amplitude is None and max_ratio is None and max_dB is None)
        
        self.max_amplitude = max_amplitude
        self.max_ratio = max_ratio
        self.max_dB = max_dB
        self.min_amplitude = min_amplitude
        self.min_ratio = min_ratio
        self.min_dB = min_dB
    
    def realise(self, audio):
        # convert to the amplitude case then continue normally
        if self.is_max:
            if self.max_ratio is not None:
                self.max_amplitude = self.max_ratio*np.max(np.abs(audio.audio))
        
            # TODO do the same for dBs
            np.clip(audio.audio, -self.max_amplitude, self.max_amplitude, out=audio.audio)
        
        if self.is_min:
            if self.min_ratio is not None:
                self.min_amplitude = self.min_ratio*np.max(np.abs(audio.audio))
            #TODO same for dBs
            
            audio.audio[:,:] = np.sign(audio.audio) * \
                               np.clip(np.abs(audio.audio), a_min=self.min_amplitude, a_max=None)
        


###### PANNING STUFF

class Mono(Transform):
    """ transforms multi-channel audio to mono
    """
    def realise(self, audio):
        audio.audio = np.sum(audio.audio, axis=0, keepdims=True)
        # TODO should we normalize the sum?

class Pan(Transform):
    """ applies arbitrary function to amplitudes of all channels
    """
    width = 200 # means hard left is -100, hard right is 100
    #eps = 0.001 # can use this to avoid np.log(0) warning; not necessary
    # i.e. np.log((x+hardR)/width) below
    
    # you can set Pan.panLaw once and for all; anytime before mixdown() js good
    panLaw = -3 # -3 seems appropriate for headphones
    
    # TODO faster computations
    # TODO put the default stereo scheme as well as the pan law as package variables
    pan_shape = lambda x: np.log(x/Pan.width + 0.5)*(-Pan.panLaw / np.log(2)) # +0.1 to prevent log(0)
    LdB = lambda x: Pan.pan_shape(-x)
    RdB = lambda x: Pan.pan_shape(x)
    defaultStereo = lambda x: (Pan.LdB(x), Pan.RdB(x))
    
    def __init__(self, pan, scheme=defaultStereo):
        """ pan is a number or curve, which will be fed into the
        underlying panning scheme.
        
        The underlying panning scheme is a number or Curve/MultiCurve in R -> R^n.
        n is interpreted as number of affected output channels,
        and the output values as gains.
        the domain R may be defined by the user whichever way they want.
        
        the default panning scheme maps numbers in [-100,100] into (-inf, 0]^2
        """
        self.pan = pan
        self.scheme = scheme
    
    def realise(self, audio):
        # or maybe we can string some monos together and apply same panning for all?
        assert audio.num_channels == 1, "panning is from mono to multi"
        
        if isnumber(self.pan):
            dBs = self.scheme(self.pan)
        elif isinstance(self.pan, Curve):
            dBs = self.scheme(self.pan.flatten(audio.sample_rate))
        
        audio.from_mono(len(dBs))
        
        for (i, dB) in enumerate(dBs):
            audio.audio[i,:len(dB)] *= DB_to_Linear(dB)
            audio.audio[i,len(dB):] *= DB_to_Linear(self.scheme(self.pan.endpoint())[i])

class Repan(Transform):
    """ Allows switching between channels.
    """
    def __init__(self, *channels):
        """ channels is a permutation on the existing channels.
        i.e. (1,0) switches left and right.
        None means leave channel empty.
        """
        assert isinstance(channels, tuple)
        self.channels = channels
    
    def realise(self, audio):
        new_audio = np.zeros(audio.audio.shape, dtype=np.float64)
        for i, channel in enumerate(self.channels):
            if channel is None:
                continue
            new_audio[i,:] = audio.audio[channel,:]
            
        audio.audio[:,:] = new_audio[:,:]


####### FILTERS #########

class Stretch(Transform):
    """ Stretches audio by a certain factor, or to a desired duration,
    using interpolation.
    """
    # TODO parametric stretch!
    def __init__(self, rate=None, duration=None, method="quadratic"):
        """ exactly one of factor, duration should be defined.
        factor = ratio of speed up/slow down. (1=no difference, 2=twice as fast)
        duration = stretch/shrink to this duration (milliseconds or samples)
        """
        self.rate = rate
        self.duration = duration
        self.method = method
    
    def realise(self, audio):
        assert (self.rate is None) + (self.duration is None) == 1, "Stretch: exactly one of the arguments rate, duration, must be defined."
        from gensound.utils import get_interpolation
        interpolate = get_interpolation(self.method)
        
        if self.rate is None:
            factor = audio.duration / self.duration
        else:
            factor = self.rate
        
        audio.audio = interpolate(audio.audio, np.arange(0, audio.length-1, factor))
        


class Downsample(Transform):
    """ skips samples. can hear the effects of aliasing.
    suppose factor is 3, then this copies all 3k-th samples into
    the 3k+1-th, 3k+2-th places.
    phase is supposed to let us choose 3k+1, 3k+2 as the main one for example
    # can be interesting to put 4k on L and 4k+2 on R, see if there is stereo effect
    """
    
    def __init__(self, factor=2, phase=0):
        self.factor = factor
        self.phase = phase
        assert isinstance(phase, int) and 0 <= phase < factor
        #assert phase == 0, "not implemented"
        if phase != 0:
            raise NotImplementedError # TODO
    
    def realise(self, audio):
        l = audio.audio.shape[1] #- self.phase
        
        for i in range(1, self.factor):
            less = 0 != (l % self.factor) <= i
            audio.audio[:,i::self.factor] = audio.audio[:,0:l + (-self.factor if less else 0):self.factor]
            #audio.audio[:,i::self.factor] = audio.audio[:,self.phase:l + (-self.factor if less else 0):self.factor]


###### EXPERIMENTS ######
'''
class Convolution(Transform):
    """ is this the right name?
    squaring the signal, or multiplying by its own reverse
    """
    
    def __init__(self, forward=True, add=0, is_both_ways=False):
        self.forward = forward
        self.add = add
        self.is_both_ways = is_both_ways
    
    def realise(self, audio):
        # TODO the add works both ways to keep the signal away from zero
        # but we can also omit the sign thing, basically making the signal
        # more positive in general, and sounding less strange
        other = audio.audio[:,::(1 if self.forward else -1)]
        if self.is_both_ways:
            audio.audio[:,:] *= (self.add*np.sign(other) + other)
        else:
            audio.audio[:,:] *= (self.add + other)
'''

class Convolution(Transform):
    """ Convolves a Signal with given samples (self.response), which may be given
    either as an Audio object, a WAV filename, or np.ndarray. It is internally
    stored as the latter, though this may change.
    
    If response is mono, each channel of the signal is convolved individually.
    If the signal is mono, it is convolved with each of the response channels
    individually, yielding an output with the same number of channels as response.
    If multi-channel output is undesirable, crop the response audio to one channel
    before applying this transform.
    If neither is mono, then the audio and response must have the same number of
    channels, and they are convolved individually.
    """
    def __init__(self, response):
        assert isinstance(response, (Audio, np.ndarray, str)) # Audio, direct buffer, or filename
        # TODO: accept Signal as response?
        assert "scipy" in _supported, "Convolution: SciPy not supported"
        
        if isinstance(response, np.ndarray):
            self.response = response # TODO cache this using hash
        elif isinstance(response, Audio):
            self.response = response.audio
        else:
            from gensound.signals import WAV
            self.response = WAV(response).audio # load file directly to Audio        
        
        # TODO consider converting self.response to Audio
        
        if len(self.response.shape) == 1: # ensure dimensionality. TODO too similar to Audio.ensure_2d!
            self.response.resize((1, self.response.shape[0]))
    
    def realise(self, audio):
        assert self.response.shape[0] in (audio.audio.shape[0], 1) or audio.is_mono, \
            "Convolution channel number mismatch: convolved audios must either " \
            "match in channel number, or at least one must be mono."
        # self.response should be either mono or have the same number of channels as audio
        
        from scipy.signal import convolve, oaconvolve
        
        # TODO this is too wordy
        # TODO mode="same" possibly should be "full"
        if self.response.shape[0] == 1: # apply same mono reverb to all channels
            for i in range(audio.num_channels):
                audio.audio[i,:] = convolve(audio.audio[i,:], self.response[0,:], mode="same")
        elif audio.is_mono: # input signal gains new channels
            audio.from_mono(self.response.shape[0])
            for i in range(self.response.shape[0]):
                audio.audio[i,:] = convolve(audio.audio[i,:], self.response[i,:], mode="same")
        else: # both audio and response non-mono, same number of channels
            for i in range(self.response.shape[0]): # "Parallel Stereo"
                audio.audio[i,:] = convolve(audio.audio[i,:], self.response[i,:], mode="same")


class ADSR(Transform):
    """ applied ADSR envelope to signal
    """
    # TODO what if attack+decay+release > signal.duration?
    def __init__(self, attack, decay, sustain, release, hold=0): #hold=None?
        self.attack = attack
        self.hold = hold
        self.decay = decay
        self.sustain = sustain
        self.release = release
        
    def realise(self, audio):
        env_start = Line(0, 1, self.attack) | Constant(1, self.hold)
        env_start |= Line(1, self.sustain, duration = self.decay)
        length_start = env_start.num_samples(audio.sample_rate)
        # or maybe as one envelope by extrpolating from audio.duration?
        env_end = Line(self.sustain, 0, duration=self.release)
        length_end = env_end.num_samples(audio.sample_rate)
        
        audio.audio[:,:length_start] *= env_start.flatten(audio.sample_rate)
        audio.audio[:,length_start:-length_end] *= self.sustain
        audio.audio[:,-length_end:] *= env_end.flatten(audio.sample_rate)


















